/* Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <complex.h>
#include <pnfft.h>
#include <ipnfft.h>


/* wrappers for pfft init and cleanup */
void PNX(init) (void){
  PX(init)();
}


void PNX(cleanup) (void){
  PX(cleanup)();
}


void PNX(local_size_3d)(
    const INT *N, MPI_Comm comm_cart_2d,
    INT *local_N, INT *local_N_start,
    R *lower_border, R *upper_border
    )
{
  const int d=3;

  PNX(local_size_adv)(
      d, N, comm_cart_2d,
      local_N, local_N_start, lower_border, upper_border);
}


void PNX(init_3d)(
    PNX(plan) *ths, const INT *N,
    INT local_M,
    MPI_Comm comm_cart_2d
    )
{
  const int d=3;
  unsigned nfft_flags, fftw_flags;

  nfft_flags = PNFFT_MALLOC_X | PNFFT_MALLOC_F_HAT | PNFFT_MALLOC_F | PNFFT_FFT_INIT;
  fftw_flags = FFTW_MEASURE;

  PNX(init_adv)(
      ths, d, N, local_M,
      nfft_flags, fftw_flags, comm_cart_2d);
}


void PNX(init_nodes)(
    PNX(plan) *ths, INT local_M,
    unsigned nfft_flags, unsigned nfft_finalize_flags
    )
{
  if((nfft_finalize_flags & PNFFT_FREE_F) && (ths->f != NULL))
    PNX(free)(ths->f);
  if((nfft_finalize_flags & PNFFT_FREE_X) && (ths->x != NULL))
    PNX(free)(ths->x);

  ths->local_M = local_M;
  PNX(init_x_and_f)(ths, nfft_flags);
}


/* parallel 3dNFFT with Gaussian window function */
void PNX(trafo)(
    PNX(plan) *ths
    )
{
  /* multiplication with matrix D */
  PNX(trafo_D)(ths);

  /* multiplication with matrix F */
  PNX(trafo_F)(ths);

  /* multiplication with matrix B */
  PNX(trafo_B)(ths);
}


void PNX(trafo_timed)(
    PNX(plan) *ths, double *time
    )
{
  /* multiplication with matrix D */
  time[0] = -MPI_Wtime();
  PNX(trafo_D)(ths);
  time[0] += MPI_Wtime();


  /* multiplication with matrix F */
  time[1] = -MPI_Wtime();
  PNX(trafo_F)(ths);
  time[1] += MPI_Wtime();

  /* multiplication with matrix B */
  time[2] = -MPI_Wtime();
  PNX(trafo_B)(ths);
  time[2] += MPI_Wtime();
}


void PNX(adj)(
    PNX(plan) *ths
    )
{
  /* multiplication with matrix B^T */
  PNX(adjoint_B)(ths);

  /* multiplication with matrix F^H */
  PNX(adjoint_F)(ths);

  /* multiplication with matrix D */
  PNX(adjoint_D)(ths);
}


void PNX(adj_timed)(
    PNX(plan) *ths, double *time
    )
{
  /* multiplication with matrix B^T */
  time[2] = -MPI_Wtime();
  PNX(adjoint_B)(ths);
  time[2] += MPI_Wtime();

  /* multiplication with matrix F^H */
  time[1] = -MPI_Wtime();
  PNX(adjoint_F)(ths);
  time[1] += MPI_Wtime();

  /* multiplication with matrix D */
  time[0] = -MPI_Wtime();
  PNX(adjoint_D)(ths);
  time[0] += MPI_Wtime();
}


void PNX(finalize)(
    PNX(plan) *ths, unsigned nfft_finalize_flags
    )
{
  if((nfft_finalize_flags & PNFFT_FREE_F_HAT) && (ths->f_hat != NULL))
    PNX(free)(ths->f_hat);
  if((nfft_finalize_flags & PNFFT_FREE_F) && (ths->f != NULL))
    PNX(free)(ths->f);
  if((nfft_finalize_flags & PNFFT_FREE_X) && (ths->x != NULL))
    PNX(free)(ths->x);

  if(ths->nfft_flags & PNFFT_FFT_INIT)
  {
    PX(destroy_plan)(ths->pfft_forw);
    PX(destroy_plan)(ths->pfft_back);

    PNX(free)(ths->g1);
    if(ths->nfft_flags & PNFFT_FFT_OUT_OF_PLACE)
      PNX(free)(ths->g2);
  }

  PNX(free)(ths->sigma);
  PNX(free)(ths->n);
  PNX(free)(ths->no);
  PNX(free)(ths->N);

  MPI_Comm_free(&(ths->comm_cart_2d));

  /* finalize window specific parameters */
  PNX(free)(ths->b);
//  if(ths->nfft_flags & PNFFT_FG_PSI)
    PNX(free)(ths->exp_const);
}


void PNX(init_f_hat_3d)(
    const INT *N, const INT *local_N, const INT *local_N_start, MPI_Comm comm,
    C *data
    )
{
  PX(init_input_3d)(
      N, local_N, local_N_start, comm,
      data);
}


static double random_number_less_than_one(
    void
    )
{
  double tmp;
  
  do
    tmp = ((double)rand())/RAND_MAX;
  while(tmp==1.0);
  
  return tmp;
}


void PNX(init_x_3d)(
    double *lower_border, double *upper_border, ptrdiff_t local_M,
    double *x
    )
{
  double tmp;
  
  for (ptrdiff_t j=0; j<local_M; j++){
    for(int t=0; t<3; t++){
      tmp = random_number_less_than_one();
      x[3*j+t] = (upper_border[t]-lower_border[t]) * tmp + lower_border[t];
    }
  }
}


static void print_complex_vector(
    double *data, ptrdiff_t N
    )
{
  for(ptrdiff_t l=0; l<N; l++){
    if(l%4 == 0)
      printf("\n%4td.", l/4);
    printf(" %.2e+%.2ei,", data[2*l], data[2*l+1]);
  }
  printf("\n");
}


void PNX(vpr_complex)(
    fftw_complex *data, ptrdiff_t N,
    const char *name, MPI_Comm comm
    )
{
  int size, myrank;

  if(N < 1)
    return;

  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &myrank);
  
  fflush(stdout);
  MPI_Barrier(comm);
  for(int t=0; t<size; t++){
    if(t==myrank){
      printf("\nRank %d, %s", myrank, name);
      print_complex_vector((double*) data, N);
      fflush(stdout);
    }
    MPI_Barrier(comm);
  }
}


void PNX(vpr_double)(
    double *data, ptrdiff_t N,
    const char *name, MPI_Comm comm
    )
{
  int size, myrank;

  if(N < 1)
    return;
  
  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &myrank);
  
  fflush(stdout);
  MPI_Barrier(comm);
  for(int t=0; t<size; t++){
    if(t==myrank){
      printf("\nRank %d, %s", myrank, name);
      for(ptrdiff_t l=0; l<N; l++){
        if(l%8 == 0)
          printf("\n%4td.", l/8);
        printf(" %e,", data[l]);
      }
      printf("\n");
      fflush(stdout);
    }
    MPI_Barrier(comm);
  }
}


void PNX(apr_complex_3d)(
     fftw_complex *data, INT *local_N, INT *local_N_start,
     char *name, MPI_Comm comm
     )
{
  PX(apr_complex_3d)(
     data, local_N, local_N_start, name, comm);
}
