/* Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <complex.h>
#include <pnfft.h>
#include <ipnfft.h>


static void get_pfft_permutation(
    PX(permutation) *perm);
static void pre_tensor_gaussian_permuted(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *exp_const, const PX(permutation) *perm,
    R *fg_psi);
static void convolution_with_gaussian(
    C *f_hat,
    const INT *N, const INT *n,
    const INT *local_n, const INT *local_n_start,
    const R *b,
    C *g_hat);
static void get_size_gcells(
    int m, int cutoff,
    INT *gcells_below, INT *gcells_above);
static void project_node_to_grid_permuted(
    const INT *n, int m, const R *x,
    const PX(permutation) *perm,
    INT *u_j_perm);
static void project_node_to_grid(
    const INT *n, int m, const R *x,
    INT *u_j);
static void sort_nodes_for_better_cache_handle_permuted(
    const INT *n, int m, INT local_M, const R *x,
    const PX(permutation) *perm,
    INT *sort);
static int compare_INT(
    const void* a, const void* b);
static void local_array_size_permuted(
    const INT *local_n, const INT *gcells_below, const INT *gcells_above,
    const PX(permutation) *perm,
    INT *local_ngc_perm);
static void lowest_summation_index_permuted(
    const INT *n, const INT *no, int m, const R *x, const INT *local_n_start,
    const INT *gcells_below, const PX(permutation) *perm,
    INT *u_j_perm);



void PNX(local_size_D)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart_2d,
    INT *local_N, INT *local_N_start
    )
{
  INT local_no[3], local_no_start[3];

  PNX(local_size_internal)(N, n, no, comm_cart_2d,
      local_N, local_N_start, local_no, local_no_start);
}


void PNX(local_size_B)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart_2d,
    INT *local_no, INT *local_no_start
    )
{
  INT local_N[3], local_N_start[3];

  PNX(local_size_internal)(N, n, no, comm_cart_2d,
      local_N, local_N_start, local_no, local_no_start);
}


void PNX(local_size_internal)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart_2d,
    INT *local_N, INT *local_N_start,
    INT *local_no, INT *local_no_start
    )
{
  INT alloc_local_data_in, alloc_local_gc, howmany = 1;

  alloc_local_gc = PX(local_size_3d_many)(
      n, N, no, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
      PFFT_NO_GCELLS, PFFT_NO_GCELLS, comm_cart_2d, PFFT_TRANSPOSED,
      local_N, local_N_start, local_no, local_no_start, &alloc_local_data_in);
}


void PNX(init_help)(
    PNX(plan) *ths
    )
{
  INT alloc_local_data_in, alloc_local_gc, alloc_local_f_hat, howmany = 1;
  INT gcells_below[3], gcells_above[3];
  INT local_N[3], local_N_start[3], local_no[3], local_no_start[3];

  ths->cutoff = 2*ths->m+2;
  ths->N_total = ths->n_total = 1;
  for(int t=0; t<ths->d; t++){
    ths->N_total *= ths->N[t];
    ths->n_total *= ths->n[t];
  }
  ths->sigma = (R*) PNX(malloc)(ths->d * sizeof(R), ths->comm_cart_2d);
  for(int t = 0;t < ths->d; t++)
    ths->sigma[t] = ((double)ths->n[t])/ths->N[t];


  /* Precompute Flags not yet implemented */
  if(ths->nfft_flags & (PNFFT_PRE_ONE_PSI| PNFFT_PRE_PHI_HUT)){
    PNX(message)("!!! Warning: Precompute Flags are not yet implemented in PNFFT !!!\n", ths->comm_cart_2d);
    ths->nfft_flags &= (~PNFFT_PRE_ONE_PSI); /* needed for correct nfft_finalize */
    ths->nfft_flags &= (~PNFFT_PRE_PHI_HUT); /* needed for correct nfft_finalize */
  }

  /* init window specific parameters */
  ths->b = (R*) PNX(malloc)(ths->d*sizeof(R), ths->comm_cart_2d);
  for(int t=0; t<ths->d; t++)
    ths->b[t]= 2*ths->sigma[t] / (2*ths->sigma[t]-1) * ((R)ths->m / M_PI);

  ths->exp_const = (R*) PNX(malloc)(ths->d * ths->cutoff * sizeof(R), ths->comm_cart_2d);
  for(int t=0; t<ths->d; t++)
    for(int s=0; s<ths->cutoff; s++)
      ths->exp_const[ths->cutoff*t+s] = exp(-s*s/ths->b[t])/(sqrt(M_PI*ths->b[t]));

  get_size_gcells(ths->m, ths->cutoff,
      gcells_below, gcells_above);

  alloc_local_gc = PX(local_size_3d_many)(
      ths->n, ths->N, ths->no, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
      gcells_below, gcells_above, ths->comm_cart_2d, PFFT_TRANSPOSED,
      local_N, local_N_start, local_no, local_no_start, &alloc_local_data_in);

  alloc_local_f_hat = PNX(prod_ptrdiff_t)(3, local_N);
  if(ths->nfft_flags & PNFFT_MALLOC_F_HAT)
    ths->f_hat = (C*) PNX(malloc)(sizeof(C)*alloc_local_f_hat, ths->comm_cart_2d);

  if(ths->nfft_flags & PNFFT_FFT_INIT){
    if(ths->nfft_flags & PNFFT_FFT_OUT_OF_PLACE){
      ths->g1 = (C*) PNX(malloc)(sizeof(C)*alloc_local_data_in, ths->comm_cart_2d);
      ths->g2 = (C*) PNX(malloc)(sizeof(C)*alloc_local_gc, ths->comm_cart_2d);
    }
    else{
      ths->g1 = (C*) PNX(malloc)(sizeof(C)*alloc_local_gc, ths->comm_cart_2d);
      ths->g2 = ths->g1;
    }

    ths->pfft_forw = PX(plan_many_dft_3d)(
        ths->n, ths->N, ths->no,
        howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
        gcells_below, gcells_above, ths->g1, ths->g2, ths->comm_cart_2d,
        FFTW_FORWARD, PFFT_TRANSPOSED | PFFT_FORWARD, ths->fftw_flags);

    ths->pfft_back = PX(plan_many_dft_3d)(
        ths->n, ths->no, ths->N,
        howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
        gcells_below, gcells_above, ths->g2, ths->g1, ths->comm_cart_2d,
        FFTW_FORWARD, PFFT_TRANSPOSED | PFFT_BACKWARD, ths->fftw_flags);
  }
  else {
    ths->g1 = ths->g2 = NULL;
    ths->pfft_forw = ths->pfft_back = NULL;
  }

  PNX(init_x_and_f)(ths, ths->nfft_flags);
}


void PNX(init_x_and_f)(
    PNX(plan) *ths, unsigned nfft_flags
    )
{
  if(nfft_flags & PNFFT_MALLOC_F)
      ths->f = (ths->local_M) ?
               (C*) PNX(malloc)(sizeof(C)*ths->local_M, ths->comm_cart_2d) :
               NULL;
  if(nfft_flags & PNFFT_MALLOC_X)
    ths->x = (ths->local_M) ?
             (R*) PNX(malloc)(sizeof(R)*ths->d*ths->local_M, ths->comm_cart_2d) :
             NULL;
}


void PNX(node_borders)(
    const INT *n, const INT *no,
    const INT* local_no, const INT *local_no_start,
    R *lo, R *up
    )
{
  INT tmp, local_no_end[3];

  PNX(vadd_ptrdiff_t)(3, local_no_start, local_no, local_no_end);
  
  for(int t=0; t<3; t++){
    tmp = local_no_start[t];
    if( no[t]/2 <= tmp )
      tmp +=  n[t] - no[t];
    lo[t] = (R) tmp / n[t];
    
    tmp = local_no_end[t];
    if( no[t]/2 <= tmp )
      tmp +=  n[t] - no[t];
    up[t] = (R) tmp / n[t];
  }
}


void PNX(trafo_D)(
    PNX(plan) *ths
    )
{
  INT local_N[3], local_N_start[3];
  
  PNX(local_size_D)(ths->N, ths->n, ths->no, ths->comm_cart_2d,
      local_N, local_N_start);

  convolution_with_gaussian(
      ths->f_hat, ths->N, ths->n, local_N, local_N_start, ths->b,
      ths->g1);
}


void PNX(adjoint_D)(
    PNX(plan) *ths
    )
{
  INT local_N[3], local_N_start[3];
  
  PNX(local_size_D)(ths->N, ths->n, ths->no, ths->comm_cart_2d,
      local_N, local_N_start);
  
  convolution_with_gaussian(
      ths->g1, ths->N, ths->n, local_N, local_N_start, ths->b,
      ths->f_hat);
}


static void convolution_with_gaussian(
    C *f_hat,
    const INT *N, const INT *n,
    const INT *local_N, const INT *local_N_start,
    const R *b,
    C *g_hat
)
{
  INT k0, k1, k2, l0, l1, l2, m, local_N_end[3];

  PNX(vadd_ptrdiff_t)(3, local_N_start, local_N, local_N_end);

  m=0;
  for(k0=local_N_start[0]; k0<local_N_end[0]; k0++){
    l0 = MACRO_FFTSHIFT(k0, N[0]);
    for(k1=local_N_start[1]; k1<local_N_end[1]; k1++){
      l1 = MACRO_FFTSHIFT(k1, N[1]);
      for(k2=local_N_start[2]; k2<local_N_end[2]; k2++, m++){
        l2 = MACRO_FFTSHIFT(k2, N[2]);
        g_hat[m] = MACRO_INV_PHI_HAT(l0, n[0], b[0]) * MACRO_INV_PHI_HAT(l1, n[1], b[1]) *
            MACRO_INV_PHI_HAT(l2, n[2], b[2]) * f_hat[m];
      }
    }
  }
}


void PNX(trafo_F)(
    PNX(plan) *ths
    )
{
  PX(execute)(ths->pfft_forw);
}

void PNX(adjoint_F)(
    PNX(plan) *ths
    )
{
  PX(execute)(ths->pfft_back);
}

void PNX(trafo_B)(
    PNX(plan) *ths
    )
{
  const int cutoff = ths->cutoff;
  INT j, k0, k1, k2, m0, m1, m2, pbc;
//   INT *sorted_index = NULL;
  INT local_no[3], local_no_start[3];
  INT gcells_below[3], gcells_above[3];
  INT u_j_perm[3], local_ngc_perm[3];
  R *pre_psi_perm = NULL;
  PX(permutation) perm;
  
  get_pfft_permutation(&perm);
  
  PNX(local_size_B)(ths->N, ths->n, ths->no, ths->comm_cart_2d,
      local_no, local_no_start);
  
  get_size_gcells(ths->m, ths->cutoff,
      gcells_below, gcells_above);
  local_array_size_permuted(local_no, gcells_below, gcells_above, &perm,
      local_ngc_perm);

  /* send ghost cells in ring */
  PX(exchange_ghostcells)(ths->pfft_forw);

  /* sort indices for better cache handling */
//   sorted_index = (INT*) PNX(malloc)(sizeof(INT)*2*ths->local_M, ths->comm_cart_2d);
//   sort_nodes_for_better_cache_handle_permuted(
//       ths->n, ths->m, ths->local_M, ths->x, &perm,
//       sorted_index);

  pre_psi_perm = (R*) PNX(malloc)(sizeof(R)*cutoff*3, ths->comm_cart_2d);
  for(INT p=0; p<ths->local_M; p++){
    j = p;
//     j = sorted_index[2*p+1];

    /* evaluate Gaussian window on axes */
    pre_tensor_gaussian_permuted(
        ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]), ths->exp_const, &perm,
        pre_psi_perm);

    lowest_summation_index_permuted(
        ths->n, ths->no, ths->m, &(ths->x[ths->d*j]), local_no_start, gcells_below, &perm,
        u_j_perm);
    
    ths->f[j] = 0;
    m0 = MACRO_PLAIN_INDEX_3D(u_j_perm, local_ngc_perm);
    for(k0=0; k0<cutoff; k0++, m0 += local_ngc_perm[1]*local_ngc_perm[2]){
      for(k1=0, m1=m0; k1<cutoff; k1++, m1 += local_ngc_perm[2]){
        for(k2=0, m2 = m1, pbc=u_j_perm[2]; k2<cutoff; k2++, m2++, pbc++){
          if(pbc == local_ngc_perm[2]){
            m2 -= local_ngc_perm[2];
            pbc = 0;
          }
          ths->f[j] += pre_psi_perm[k0] * pre_psi_perm[cutoff+k1] * pre_psi_perm[2*cutoff+k2] * ths->g2[m2];
        }
      }
    }
  }
//  PNX(free)(sorted_index);
  PNX(free)(pre_psi_perm);
}


void PNX(adjoint_B)(
    PNX(plan) *ths
    )
{
  const int cutoff = ths->cutoff;
  INT j, k0, k1, k2, m0, m1, m2, pbc;
//   INT *sorted_index = NULL;
  INT local_no[3], local_no_start[3];
  INT gcells_below[3], gcells_above[3];
  INT u_j_perm[3], local_ngc_perm[3], local_ngc_total;
  R *pre_psi_perm = NULL;
  PX(permutation) perm;
  
  get_pfft_permutation(&perm);

  PNX(local_size_B)(ths->N, ths->n, ths->no, ths->comm_cart_2d,
      local_no, local_no_start);
  
  get_size_gcells(ths->m, ths->cutoff,
      gcells_below, gcells_above);
  local_array_size_permuted(local_no, gcells_below, gcells_above, &perm,
      local_ngc_perm);

  local_ngc_total = PNX(prod_ptrdiff_t)(3, local_ngc_perm);
  memset(ths->g2, 0, local_ngc_total * sizeof(C));

  /* sort indices for better cache handling */
//  sorted_index = (INT*) PNX(malloc)(sizeof(INT)*2*ths->local_M, ths->comm_cart_2d);
//  sort_nodes_for_better_cache_handle_permuted(
//      ths->n, ths->m, ths->local_M, ths->x, &perm,
//      sorted_index);

  pre_psi_perm = (R*) PNX(malloc)(sizeof(R)*cutoff*3, ths->comm_cart_2d);
  for(INT p=0; p<ths->local_M; p++){
    j = p;
//     j = sorted_index[2*p+1];

    /* evaluate Gaussian window on axes */
    pre_tensor_gaussian_permuted(
        ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]), ths->exp_const, &perm,
        pre_psi_perm);

    lowest_summation_index_permuted(
        ths->n, ths->no, ths->m, &(ths->x[ths->d*j]), local_no_start, gcells_below, &perm,
        u_j_perm);

    m0 = MACRO_PLAIN_INDEX_3D(u_j_perm, local_ngc_perm);
    for(k0=0; k0<cutoff; k0++, m0 += local_ngc_perm[1]*local_ngc_perm[2]){
      for(k1=0, m1=m0; k1<cutoff; k1++, m1 += local_ngc_perm[2]){
        for(k2=0, m2 = m1, pbc=u_j_perm[2]; k2<cutoff; k2++, m2++, pbc++){
          if(pbc == local_ngc_perm[2]){
            m2 -= local_ngc_perm[2];
            pbc = 0;
          }
           ths->g2[m2] += pre_psi_perm[k0] * pre_psi_perm[cutoff+k1] * pre_psi_perm[2*cutoff+k2] * ths->f[j];
        }
      }
    }
  }
  
  /* reduce ghost cells in ring */
  PX(reduce_ghostcells)(ths->pfft_forw);
  
//   PNX(free)(sorted_index);
  PNX(free)(pre_psi_perm);
}


static void get_size_gcells(
    int m, int cutoff,
    INT *gcells_below, INT *gcells_above
    )
{
  /* since we use transposed FFT, 1st dim is local  */
  gcells_below[0] = gcells_above[0] = 0;

  /* The other dims are distributed => need for ghostcells  */
  for(int t=1; t<3; t++){
    gcells_below[t] = m;
    gcells_above[t] = cutoff - gcells_below[t] - 1;
  }

}


static void pre_tensor_gaussian_permuted(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *exp_const, const PX(permutation) *perm,
    R *fg_psi
    )
{
  const int d=3;
  int t;
  R u_j, exp_sqr, exp_lin, tmp;

  for(int p=0; p<d; p++){
    t = perm->n[p];
    u_j = (INT) floor( n[t]*x[t] - m );
    exp_sqr = exp( -SQR( n[t]*x[t]-u_j ) / b[t] );
    exp_lin = exp( 2*( n[t]*x[t]-u_j ) / b[t] );

    tmp = exp_sqr;
    for(int s=0; s<cutoff; s++){
      fg_psi[cutoff*p+s] = tmp * exp_const[cutoff*t+s];
      tmp *= exp_lin;
    }
  }
}


static void get_pfft_permutation(
    PX(permutation) *perm
    )
{
  PX(profile) prf;
  
  PX(init_default_profile)(PFFT_TRANSPOSED | PFFT_BACKWARD,
      &prf, perm);
}


static void local_array_size_permuted(
    const INT *local_n, const INT *gcells_below, const INT *gcells_above,
    const PX(permutation) *perm,
    INT *local_ngc_perm
    )
{
  int t;
  
  for(int p=0; p<3; p++){
    t = perm->n[p];
    local_ngc_perm[p] = local_n[t] + gcells_below[t] + gcells_above[t];
  }
}


static void sort_nodes_for_better_cache_handle_permuted(
    const INT *n, int m, INT local_M, const R *x,
    const PX(permutation) *perm,
    INT *sort
    )
{ /* sort must be of length 2*local_M */
  int d = 3;
  INT n_perm[3], u_j_perm[3];

  if(local_M == 0)
    return;

  for(int t=0; t<3; t++)
    n_perm[t] = n[perm->n[t]];
  
  for(INT k = 0; k < local_M; k++){
    project_node_to_grid_permuted(n, m, x+d*k, perm, u_j_perm);
    sort[2*k]   = MACRO_PLAIN_INDEX_3D(u_j_perm, n_perm);
    sort[2*k+1] = k;
  }
  qsort(sort, local_M, 2*sizeof(INT), compare_INT);
}


static int compare_INT(
    const void* a, const void* b
    )
{
  return (*(INT*)a <= *(INT*)b) ? -1 : 1;
}


static void lowest_summation_index_permuted(
    const INT *n, const INT *no, int m, const R *x, const INT *local_no_start,
    const INT *gcells_below, const PX(permutation) *perm,
    INT *u_j_perm
    )
{
  int t;
  INT u_j[3], local_n_start[3];
  
  for(int t=0; t<3; t++){
    local_n_start[t] = local_no_start[t];
    if( no[t]/2 <= local_no_start[t] )
      local_n_start[t] +=  n[t] - no[t];
  }
  
  project_node_to_grid(n, m, x, u_j);
  for(int p=0; p<3; p++){
    t = perm->n[p];
    u_j_perm[p] = u_j[t] -local_n_start[t] + gcells_below[t];

    /* last dim is local in memory and has no ghostcells */
    /* assure periodic boundary */
    if(p==2)
      u_j_perm[2] = (u_j_perm[2]%n[t]+n[t])%n[t];
  }
}


static void project_node_to_grid_permuted(
    const INT *n, int m, const R *x,
    const PX(permutation) *perm,
    INT *u_j_perm
    )
{
  INT u_j[3];
  
  project_node_to_grid(n, m, x, u_j);
  for(int t=0; t<3; t++)
    u_j_perm[t] = u_j[perm->n[t]];
}


static void project_node_to_grid(
    const INT *n, int m, const R *x,
    INT *u_j
    )
{
  for(int t=0; t<3; t++)
    u_j[t] = (INT) floor(n[t]*x[t]) - m;
}

