/*
 * Copyright (c) 2010 Michael Pippig
 *
 * The following statement of license applies *only* to this header file,
 * and *not* to the other files distributed with FFTW or derived therefrom:
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef PNFFT_H
#define PNFFT_H

#include <math.h>
#include <fftw3.h>
#include <pfft.h>
#include <mpi.h>

#ifdef __cplusplus
  extern "C"
  {
#endif /* __cplusplus */


#define PNFFT_EXTERN extern
#define PNFFT_CONCAT(prefix, name) prefix ## name

  /*
    huge second-order macro that defines prototypes for all API
    functions.  We expand this macro for each supported precision

    PNX: PNFFT name-mangling macro (parallel NFFT)
    PX: PFFT name-mangling macro (parallel FFT)
    X : FFTW name-mangling macro (serial FFT)
    R : real data type
    C : complex data type
  */
#define PNFFT_DEFINE_API(PNX, PX, X, R, C)                                              \
                                                                                        \
  typedef struct{                                                                       \
    ptrdiff_t N_total;          /**< Total number of Fourier coefficients            */ \
    C *f_hat;                   /**< Vector of Fourier coefficients                  */ \
    C *f;                       /**< Vector of samples                               */ \
    R *x;                       /**< Nodes in time/spatial domain                    */ \
                                                                                        \
    int d;                      /**< Dimension, rank                                 */ \
    ptrdiff_t *N;               /**< Multi bandwidth                                 */ \
    R *sigma;                   /**< Oversampling-factor                             */ \
    ptrdiff_t *n;               /**< FFT length, equal to sigma*N                    */ \
    ptrdiff_t *no;              /**< FFT output length                               */ \
    ptrdiff_t n_total;          /**< Total size of FFTW                              */ \
    int m;                      /**< Cut-off parameter of the window function        */ \
                                                                                        \
    R *b;                       /**< Shape parameter of Gaussian window function     */ \
    R *exp_const;               /**< Precomputed values for Fast Gaussian window     */ \
                                                                                        \
    unsigned nfft_flags;        /**< Flags for precomputation, (de)allocation,          \
                                     and FFTW usage                                  */ \
    unsigned fftw_flags;        /**< Flags for the FFTW                              */ \
                                                                                        \
    /* internal*/                                                                       \
    PX(plan)  pfft_forw;        /**< Forward PFFT plan                               */ \
    PX(plan)  pfft_back;        /**< Backward PFFT plan                              */ \
                                                                                        \
    C *g1;                      /**< Input of PFFT                                   */ \
    C *g2;                      /**< Output of PFFT                                  */ \
                                                                                        \
    int cutoff;                 /**< cutoff range                                    */ \
    ptrdiff_t local_M;          /**< Number of local nodes                           */ \
                                                                                        \
    MPI_Comm comm_cart_2d;      /**< Two-dimensional Cartesian communicator          */ \
  } PNX(plan);                                                                          \
                                                                                        \
  PNFFT_EXTERN int PNX(create_procmesh_2d)(                                             \
      MPI_Comm comm, int np0, int np1, MPI_Comm *comm_cart_2d);                         \
                                                                                        \
  PNFFT_EXTERN void PNX(local_size_3d)(                                                 \
      const ptrdiff_t *N, MPI_Comm comm_cart_2d,                                        \
      ptrdiff_t *local_N, ptrdiff_t *local_N_start,                                     \
      R *lower_border, R *upper_border);                                                \
  PNFFT_EXTERN void PNX(local_size_adv)(                                                \
      int d, const ptrdiff_t *N, MPI_Comm comm_cart_2d,                                 \
      ptrdiff_t *local_N, ptrdiff_t *local_N_start,                                     \
      R *lower_border, R *upper_border);                                                \
  PNFFT_EXTERN void PNX(local_size_guru)(                                               \
      int d, const ptrdiff_t *N, const ptrdiff_t *n, const R *x_max, int m,             \
      MPI_Comm comm_cart_2d,                                                            \
      ptrdiff_t *local_N, ptrdiff_t *local_N_start,                                     \
      R *lower_border, R *upper_border);                                                \
                                                                                        \
  PNFFT_EXTERN void PNX(init_3d)(                                                       \
      PNX(plan) *ths, const ptrdiff_t *N, ptrdiff_t local_M,                            \
      MPI_Comm comm_cart_2d);                                                           \
  PNFFT_EXTERN void PNX(init_adv)(                                                      \
      PNX(plan) *ths, int d, const ptrdiff_t *N, ptrdiff_t local_M,                     \
      unsigned nfft_flags, unsigned fftw_flags, MPI_Comm comm_cart_2d);                 \
  PNFFT_EXTERN void PNX(init_guru)(                                                     \
        PNX(plan) *ths, int d,                                                          \
        const ptrdiff_t *N, const ptrdiff_t *n, const R *x_max,                         \
        ptrdiff_t local_M, int m,                                                       \
        unsigned nfft_flags, unsigned fftw_flags,                                       \
        MPI_Comm comm_cart_2d);                                                         \
                                                                                        \
  PNFFT_EXTERN void PNX(init_nodes)(                                                    \
      PNX(plan) *ths, ptrdiff_t local_M,                                                \
      unsigned nfft_flags, unsigned nfft_finalize_flags);                               \
                                                                                        \
  PNFFT_EXTERN void PNX(finalize)(                                                      \
        PNX(plan) *ths, unsigned nfft_finalize_flags);                                  \
                                                                                        \
  PNFFT_EXTERN void PNX(trafo)(                                                         \
      PNX(plan) *ths);                                                                  \
  PNFFT_EXTERN void PNX(trafo_timed)(                                                   \
      PNX(plan) *ths, double *time);                                                    \
  PNFFT_EXTERN void PNX(adj)(                                                           \
      PNX(plan) *ths);                                                                  \
  PNFFT_EXTERN void PNX(adj_timed)(                                                     \
      PNX(plan) *ths, double *time);                                                    \
                                                                                        \
  PNFFT_EXTERN void *PNX(malloc)(                                                       \
      size_t n, MPI_Comm comm);                                                         \
  PNFFT_EXTERN void PNX(free)(                                                          \
      void *p);                                                                         \
                                                                                        \
  PNFFT_EXTERN void PNX(init)(                                                          \
      void);                                                                            \
  PNFFT_EXTERN void PNX(cleanup)(                                                       \
      void);                                                                            \
  void PNX(init_f_hat_3d)(                                                              \
      const ptrdiff_t *N, const ptrdiff_t *local_N,                                     \
      const ptrdiff_t *local_N_start, MPI_Comm comm,                                    \
      C *data);                                                                         \
  void PNX(init_x_3d)(                                                                  \
    double *lower_border, double *upper_border, ptrdiff_t local_M,                      \
    double *x);                                                                         \
                                                                                        \
  void PNX(vpr_complex)(                                                                \
    fftw_complex *data, ptrdiff_t N, const char *name, MPI_Comm comm);                  \
  void PNX(vpr_double)(                                                                 \
    double *data, ptrdiff_t N, const char *name, MPI_Comm comm);                        \
  void PNX(apr_complex_3d)(                                                             \
    fftw_complex *data, ptrdiff_t *local_N, ptrdiff_t *local_N_start,                   \
    char *name, MPI_Comm comm);












#define PNFFT_MANGLE_DOUBLE(name) PNFFT_CONCAT(pnfft_, name)
#define PNFFT_MANGLE_FLOAT(name) PNFFT_CONCAT(pnfftf_, name)
#define PNFFT_MANGLE_LONG_DOUBLE(name) PNFFT_CONCAT(pnfftl_, name)

PNFFT_DEFINE_API(
    PNFFT_MANGLE_DOUBLE, PFFT_MANGLE_DOUBLE, FFTW_MANGLE_DOUBLE,
    double, fftw_complex)
PNFFT_DEFINE_API(
    PNFFT_MANGLE_FLOAT, PFFT_MANGLE_FLOAT, FFTW_MANGLE_FLOAT,
    float, fftwf_complex)
PNFFT_DEFINE_API(
    PNFFT_MANGLE_LONG_DOUBLE, PFFT_MANGLE_LONG_DOUBLE, FFTW_MANGLE_LONG_DOUBLE,
    long double, fftwl_complex)



#ifndef M_PI
#define M_PI           3.14159265358979323846
#endif

#define PNFFT_PRE_PHI_HUT      (1U<< 0)
#define PNFFT_FG_PSI           (1U<< 1)
#define PNFFT_PRE_LIN_PSI      (1U<< 2)
#define PNFFT_PRE_FG_PSI       (1U<< 3)
#define PNFFT_PRE_PSI          (1U<< 4)
#define PNFFT_PRE_FULL_PSI     (1U<< 5)

#define PNFFT_MALLOC_X         (1U<< 6)
#define PNFFT_MALLOC_F_HAT     (1U<< 7)
#define PNFFT_MALLOC_F         (1U<< 8)
#define PNFFT_FFT_OUT_OF_PLACE (1U<< 9)
#define PNFFT_FFT_INIT         (1U<< 10)

#define PNFFT_PRE_ONE_PSI (PNFFT_PRE_LIN_PSI| PNFFT_PRE_FG_PSI| PNFFT_PRE_PSI| PNFFT_PRE_FULL_PSI)

#define PNFFT_FREE_X         PNFFT_MALLOC_X
#define PNFFT_FREE_F_HAT     PNFFT_MALLOC_F_HAT
#define PNFFT_FREE_F         PNFFT_MALLOC_F



#ifdef __cplusplus
  }  /* extern "C" */
#endif /* __cplusplus */

#endif /* PNFFT_H */

